package daltonism;

import java.io.*;
import javafx.beans.property.*;
import javafx.beans.InvalidationListener;
import javafx.application.Application;
import javafx.scene.Scene;
import javafx.scene.layout.*;
import javafx.scene.text.*;
import javafx.scene.Node;
import javafx.stage.Stage;
import java.nio.IntBuffer;
import javafx.scene.image.*;
import javafx.scene.control.*;
import javafx.geometry.*;
import javafx.scene.control.*;
import javafx.scene.control.Alert.*;
import static daltonism.Anomaly.*;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;
import javafx.scene.image.WritableImage;
import javafx.event.*;
import javafx.embed.swing.SwingFXUtils;
import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import javafx.scene.paint.*;

public class Daltoniser extends Application {
  static String[] arguments;
  
  Stage primaryStage;
  AnomalyImage anomalyImage = new AnomalyImage();
  PixelReader origpxReader;
  byte[] origBuffer;
  File fileOrig = null;
  ImageView imageView;
  Pane imgPane = new Pane();
  ScrollPane scrollPane = new ScrollPane();
  BorderPane rootPane = new BorderPane();
  HBox bottom = new HBox();
  Slider sl = new Slider();
  Region spaceSl = new Region();
  Button loadBtn = new Button("Load");
  Button saveBtn = new Button("Save");
  Button aboutBtn = new Button("About");
  VBox left = new VBox();
  GridPane btnsPane = new GridPane();
  RadioButton deutBtn = new RadioButton("Deuteranomaly");
  RadioButton propBtn = new RadioButton("Protanomaly");
  RadioButton tritBtn = new RadioButton("Tritanomaly");
  CheckBox reverseCb = new CheckBox("Reverse");
  Text explanationTxt = new Text("Click for help");
  CheckBox alwaysUpdateCb = new CheckBox("Update Continuously");
  CheckBox fitToWinCb = new CheckBox("Fit to windows");  
  InvalidationListener centerResize = ov -> centerResize();
  // For the slider
  InvalidationListener processl = ov -> process();
  // For the buttons
  EventHandler<ActionEvent> processa = ov -> process();
    
  @Override
  public void start(Stage primaryStage) {
    this.primaryStage = primaryStage;
    // Set the root pane layout
    Scene scene = new Scene(rootPane,800,800);
    primaryStage.setScene(scene);
    primaryStage.setTitle("Daltoniser");
    rootPane.setCenter(scrollPane);
    rootPane.setBottom(bottom);
    bottom.setPadding(new Insets(15,5,5,10));
    rootPane.setLeft(left);
    left.setPadding(new Insets(5,5,5,5));

    // Set the center: ImageView
    scrollPane.setContent(imgPane);
    scrollPane.setPannable(true);
    scrollPane.setFitToWidth(true);
    scrollPane.setFitToHeight(true);
    
    fileOrig = null;
    boolean success = false;
    if (arguments.length != 0) {
      fileOrig = new File(arguments[0]);
      success = anomalyImage.setImage(fileOrig);
    }
    
    imageView = new ImageView();

    if (success) imageView.setImage(anomalyImage.getImage());
    imgPane.getChildren().add(imageView);
    imgPane.setStyle("-fx-background-color: black;");
    imgPane.widthProperty().addListener(centerResize);
    imgPane.heightProperty().addListener(centerResize);
    imageView.setPreserveRatio(true);

    // Set the bottom
    Separator separator = new Separator(Orientation.VERTICAL);
    bottom.getChildren().addAll(new Text("Severity: "), sl, spaceSl, separator,alwaysUpdateCb);
    HBox.setMargin(separator, new Insets(0,10,0,10));
    setSlider();
    
    // Set Left
    left.setSpacing(5);
    left.getChildren().add(btnsPane);
    // btnsPane.setGridLinesVisible(true);
    ColumnConstraints column1 = new ColumnConstraints();
    column1.setPercentWidth(50);
    ColumnConstraints column2 = new ColumnConstraints();
    column2.setPercentWidth(50);
    btnsPane.getColumnConstraints().addAll(column1, column2);
    for (Button b : new Button[] {loadBtn,saveBtn,aboutBtn}) b.setMaxWidth(Double.MAX_VALUE);
    btnsPane.setHgap(5);
    btnsPane.setVgap(5);
    left.setFillWidth(true);
    btnsPane.addRow(0,loadBtn,saveBtn);
    btnsPane.addRow(1,aboutBtn);
    explanationTxt.setUnderline(true);
    explanationTxt.setFill(Color.BLUE);
    left.getChildren().addAll(new Separator(), deutBtn, propBtn, tritBtn, new Separator(),
                              new Text("Experimental"), explanationTxt, reverseCb,
                              new Separator(), fitToWinCb);

    // Set the state of the buttons. 
    ToggleGroup group = new ToggleGroup();
    deutBtn.setToggleGroup(group);
    propBtn.setToggleGroup(group);
    tritBtn.setToggleGroup(group);
    deutBtn.setSelected(true);
    fitToWinCb.setSelected(true);
    
    // Add listeners & Actions for components
    fitToWinCb.disableProperty().bind(anomalyImage.imagePresentProperty().not());
    reverseCb.disableProperty().bind(anomalyImage.imagePresentProperty().not());
    saveBtn.disableProperty().bind(anomalyImage.imagePresentProperty().not());
    sl.disableProperty().bind(anomalyImage.imagePresentProperty().not());
    deutBtn.setOnAction(processa);
    deutBtn.disableProperty().bind(anomalyImage.imagePresentProperty().not());
    propBtn.setOnAction(processa);
    propBtn.disableProperty().bind(anomalyImage.imagePresentProperty().not());
    tritBtn.setOnAction(processa);
    tritBtn.disableProperty().bind(anomalyImage.imagePresentProperty().not());
    sl.valueProperty().addListener(processl);
    sl.valueChangingProperty().addListener(processl);
    alwaysUpdateCb.setOnAction(e -> {
        if (alwaysUpdateCb.isSelected()) sl.valueChangingProperty().removeListener(processl);
        else sl.valueChangingProperty().addListener(processl);
      });
    alwaysUpdateCb.disableProperty().bind(anomalyImage.imagePresentProperty().not());
    loadBtn.setOnAction(e -> loadFile());
    saveBtn.setOnAction(e -> saveFile());
    aboutBtn.setOnAction(e -> about());
    fitToWinCb.setOnAction(e -> refitFull());
    explanationTxt.setOnMouseClicked(e -> helpReverse());
    reverseCb.setOnAction(e -> {
        setSlider();
        process();
      });
    primaryStage.show();
  }

  void refitFull() {
    if (fitToWinCb.isSelected()) {
      imgPane.setMinWidth(0);
      imgPane.setMinHeight(0);
    } else {
      imgPane.setMinWidth(anomalyImage.getWidth());
      imgPane.setMinHeight(anomalyImage.getHeight());
    }
    centerResize();
  }
  
  void centerResize() {
    double panewidth = imgPane.getWidth();
    double paneheight = imgPane.getHeight();
    double imagewidth = anomalyImage.getWidth();
    double imageheight = anomalyImage.getHeight();
    if (fitToWinCb.isSelected()) {
      imageView.setFitWidth(Math.min(panewidth,imagewidth));
      imageView.setFitHeight(Math.min(paneheight,imageheight));
    } else {
      imageView.setFitWidth(imagewidth);
      imageView.setFitHeight(imageheight);
    }
    imageView.setX((panewidth-imageView.getBoundsInLocal().getWidth())/2);
    imageView.setY((imgPane.getHeight()-imageView.getBoundsInLocal().getHeight())/2);
  }
  
  void process() {
    // System.out.printf("deutBtn selected: %b, disabled: %b\n", deutBtn.isSelected(),deutBtn.isDisable());
    // System.out.printf("propBtn: %b\n", propBtn.isSelected());
    // System.out.printf("tritBtn: %b\n", tritBtn.isSelected());
    // System.out.printf("alwaysUpdateCb: %b\n", alwaysUpdateCb .isSelected());
    // System.out.printf("Slider: %f, Changing: %b\n",sl.getValue(),sl.isValueChanging());
    // System.out.printf("Image present: %b", anomalyImage.isImagePresent());
    // System.out.printf("\n\n");

    if (!alwaysUpdateCb.isSelected() && sl.isValueChanging()) return;
    int anomaly = -1;
    if (deutBtn.isSelected()) anomaly = DEUT;
    else if (tritBtn.isSelected()) anomaly = TRIT;
    else if (propBtn.isSelected()) anomaly = PROT;
    assert (anomaly != -1);
    float severity = (float)(sl.getValue()/100);
    anomalyImage.daltonise(anomaly, severity, reverseCb.isSelected());
  }

  void helpReverse() {
    Alert alert = new Alert(AlertType.INFORMATION);
    alert.setTitle("Reverse");
    alert.setHeaderText("Reverse anomaly");
    alert.setResizable(true);
    alert.getDialogPane().setPrefSize(400, 380);
    alert.setContentText("The reverse option applies the transformation in reverse. It is the recoloring "+
                         "that would bring the recolored image back to the original. Note that this "+
                         "only works within the limit of the color model. The severity corresponds to "+ 
                         "the value of the direct transformation; it is limited by 50. Going further "+
                         "than that would almost always go out of the color model and thus would "+
                         "make no sense.\n\n"+
                         "Doing the transformation in reverse might give a taste "+
                         "to mildly color deficient people of what normal vision "+
                         "people see. Please note that unlike the direct transformation, "+
                         "it is highly conjectural and is not supported by serious researchers. "+
                         "This should be considered as an experiment only.");
    alert.show();
  }

  void setSlider() {
    if (reverseCb.isSelected()) {
      sl.setValue(0);   
      sl.setMax(50);
      sl.setMin(0);
      sl.setMajorTickUnit(10);
      sl.setShowTickMarks(true);
      sl.setShowTickLabels(true);
      sl.setBlockIncrement(10);
      sl.setValue(0);
      sl.setPrefWidth(250);
      spaceSl.setPrefWidth(250);
    } else {
      sl.setValue(0);   
      sl.setMax(100);
      sl.setMin(0);
      sl.setMajorTickUnit(10);
      sl.setShowTickMarks(true);
      sl.setShowTickLabels(true);
      sl.setBlockIncrement(10);
      sl.setValue(0);
      sl.setPrefWidth(500);
      spaceSl.setPrefWidth(0);
    }
  }
    
    void loadFile() {
    FileChooser fileChooser = new FileChooser();
    fileChooser.setTitle("Choose image to open.");
    fileOrig = fileChooser.showOpenDialog(primaryStage);
    if (fileOrig != null) {
      if (anomalyImage.setImage(fileOrig)) {
        imageView.setImage(anomalyImage.getImage());
        // Maybe with a listener on a Property fileChanged, seems overkill for
        // this case only.
        refitFull();
        process();
      }
    }
  }

  void saveFile() throws IllegalStateException {
    if (!anomalyImage.isImagePresent())
      throw new IllegalStateException("Cannot save the image, image not present.");
    FileChooser fileChooser = new FileChooser();
    fileChooser.getExtensionFilters().addAll(
         new ExtensionFilter("JPEG", "*.jpg"),
         new ExtensionFilter("PNG", "*.png"),
         new ExtensionFilter("GIF", "*.gif"));

    assert (fileOrig != null);
    fileChooser.setInitialDirectory(fileOrig.getParentFile());
    File file = fileChooser.showSaveDialog(primaryStage);
    if (file == null) return;
    String format = fileChooser.getSelectedExtensionFilter().getExtensions().get(0).substring(2);
    // Make a BufferedImage without alpha channel (the jpeg encoder is buggy if there is an alpha channel)
    // We will contruct it from newImagePixels.
    BufferedImage  bImage = new BufferedImage(anomalyImage.width, anomalyImage.height, BufferedImage.TYPE_INT_RGB);
    bImage.setRGB(0,0,anomalyImage.width,anomalyImage.height,anomalyImage.newImagePixels,0,anomalyImage.width);
    boolean success = false;
    try {
      // success = ImageIO.write(SwingFXUtils.fromFXImage(anomalyImage.getImage(), null), format, file);
      success = ImageIO.write(bImage,format,file);
    } catch (IOException ex) {
      Alert alert = new Alert(AlertType.ERROR);
      alert.setTitle("Error");
      alert.setHeaderText("Cannot write the image.");
      alert.setContentText("Unexpected error when writing the image.");
      alert.showAndWait();
    }
    if (!success) {
      Alert alert = new Alert(AlertType.ERROR);
      alert.setTitle("Error");
      alert.setHeaderText("Cannot write the image.");
      alert.setContentText("Format unsupported.");
      alert.showAndWait();
    }
}

  void about() {
    Alert alert = new Alert(AlertType.INFORMATION);
    alert.setTitle("About?");
    alert.setHeaderText("The daltoniser (version 1.2)");
    alert.setResizable(true);
    alert.getDialogPane().setPrefSize(600, 300);
    alert.setContentText("Copyright\u00a9 Olivier Esser (2017) (olivier.esser@gmail.com).\n"+
                         "Site web: https://freeshell.de/~oesser/daltoniser\n"+
                         "\n"+
                         "This software is available under the GNU General public license version 2. "+
                         "It comes without any warranty of any kind.\n"+
                         "\n"+
                         "The algorithm used in this software is based on the research of "+
                         "Gustavo M. Machado, Manuel M. Oliveira, Leandro A. F. Fernandes "+
                         "More information can be found on their website: "+
                         "http://www.inf.ufrgs.br/~oliveira/pubs_files/CVD_Simulation/CVD_Simulation.html");
    alert.show();
  }

  public static void main(String[] args) {
    arguments=args;
    launch(args);
  }
}
  
class AnomalyImage {
  static final WritablePixelFormat<IntBuffer> intArgbFormat = PixelFormat.getIntArgbInstance();
  
  private final ReadOnlyBooleanWrapper imagePresent = new ReadOnlyBooleanWrapper(false);
  public final boolean isImagePresent() {
    return imagePresent.get();
  }
  public final ReadOnlyBooleanProperty imagePresentProperty() {
    return imagePresent.getReadOnlyProperty();
  }
  
  // Byte buffer for the original image
  private int[] origImagePixels;
  int[] newImagePixels;
  //The current image
  private WritableImage image = null;
  private PixelWriter imageWriter;
  int height,width;
  private float[][] matrix = new float[3][3];
  private Image origImage = null;

  // Inverse a 3x3 matrix
  // Currently no check is done to ensure integrity.
  private float[][] inverseMatrix(float[][] M) throws IllegalArgumentException {
    float a=M[0][0];
    float b=M[0][1];
    float c=M[0][2];
    float d=M[1][0];
    float e=M[1][1];
    float f=M[1][2];
    float g=M[2][0];
    float h=M[2][1];
    float i=M[2][2];
    float invDet=1/(a*(e*i-f*h)-b*(d*i-f*g)+c*(d*h-e*g));
    float[][] N = {{(e*i-f*h)*invDet, (c*h-b*i)*invDet, (b*f-c*e)*invDet}, {(f*g-d*i)*invDet, (a*i-c*g)*invDet, (c*d-a*f)*invDet},
                   {(d*h-e*g)*invDet, (b*g-a*h)*invDet, (a*e-b*d)*invDet}};
    return N;
  }
  
  public AnomalyImage() {
  }

  // Set the image from file. Currently, the only supported way to set an image.
  // Return true if successful and false otherwise. Shows the alerts to the user (good idea?).
  public boolean setImage(File inputImage) {
    boolean filefound = true;
    FileInputStream is = null;
    try {
      is = new FileInputStream(inputImage);
      origImage = new Image(is);
    }
    catch (FileNotFoundException ex) {
      Alert alert = new Alert(AlertType.ERROR);
      alert.setTitle("Error");
      alert.setHeaderText("File not found.");
      alert.setContentText("I cannot find the file:\n"+inputImage.getPath());
      alert.showAndWait();
      return false;
    }
    finally {
      try {
        if (is != null) is.close();
      }
      catch (IOException ioex) {
        System.out.println("Cannot close the file!");
        System.exit(1);
      }
    }
    
    PixelReader pxReader = origImage.getPixelReader();
    if (pxReader == null) {
      Alert alert = new Alert(AlertType.ERROR);
      alert.setTitle("Error");
      alert.setHeaderText("Cannot open the image.");
      alert.setContentText("Format unsupported.");
      alert.showAndWait();
      return false;
    } else {
      width = (int)origImage.getWidth();
      height = (int)origImage.getHeight();
      origImagePixels = new int[width*height];
      pxReader.getPixels(0,0,width,height,intArgbFormat,origImagePixels,0,width);
      image = new WritableImage(pxReader,width,height);
      origImage = null;
      pxReader = null;
      imageWriter = image.getPixelWriter();
      newImagePixels = new int[width*height];
      imagePresent.set(true);
      return true;
    }
  }
  
  Image getImage() { return image; }
  double getHeight() { return (double)(height); }
  double getWidth() { return (double)(width); }
  
  void daltonise(int anomaly, float fseverity, boolean reverse) throws IllegalArgumentException {
    int bottomSeverity;
    int topSeverity;
    float[][][][] MATR = AnomalyMatrices;

    if (!isImagePresent()) return; 
    if (anomaly != PROT && anomaly != DEUT && anomaly != TRIT) {
      throw new IllegalArgumentException("This does not correspond to an anomaly");
    }
    // Allow small rounding errors in the arguments.
    if (fseverity < -0.001f || fseverity > 1.001f) {
      throw new IllegalArgumentException("The severity must be beween 0 and 1");
    }
    int severity = (int)(fseverity*10);
    if (severity < 0) severity = 0;
    
    for (int lig=0; lig < 3; lig++)
      for (int col=0; col < 3; col++) {
        if (severity == 10)
          matrix[lig][col] = AnomalyMatrices[anomaly][severity][lig][col];
        else {
          float bottom = AnomalyMatrices[anomaly][severity][lig][col];
          float top = AnomalyMatrices[anomaly][severity+1][lig][col];
          matrix[lig][col] = bottom + (fseverity*10 - severity)*(top - bottom);
        }
      }

    if (reverse) matrix=inverseMatrix(matrix);

    for (int k=0; k < width*height; k++) {
      int pixel = origImagePixels[k];
      float red   = (float)((pixel >> 16) & 0xff);
      float green = (float)((pixel >>  8) & 0xff);
      float blue  = (float)((pixel      ) & 0xff);
      float color[] = {red,green,blue};
      int newpixel = (255 << 24); // Make the picture opaque
      for (int l=0, shift=16; l < 3; l++, shift-=8) {
        float sum = 0f;
        for (int c=0; c<3; c++)
          sum = sum + matrix[l][c]*color[c];
        if (sum<0); // Useless: newpixel = newpixel | (0 << shift);
        else if (sum >= 255) newpixel = newpixel | (255 << shift);
        else newpixel = newpixel | (((int)(sum+0.5f)) << shift);
      }
      newImagePixels[k] = newpixel;
    }
    imageWriter.setPixels(0,0,width,height,intArgbFormat,newImagePixels,0,width);
  }
}
  
